from swyft.inference.marginalposterior import MarginalPosterior
from swyft.inference.marginalratioestimator import MarginalRatioEstimator

__all__ = [
    "MarginalRatioEstimator",
    "MarginalPosterior",
]
